from enum import Enum
import math
import copy

from sqc.src.Operand import Operand


# Enum of possible binary operations.
class BinOp(Enum):
    Add = 1
    Sub = 2
    Mult = 3
    Div = 4
    Pow = 5
    log = 6
    mod = 7


class VectorBinOp(Enum):
    Add = 1
    Sub = 2
    MultByNum = 3
    DotProd = 4
    GetInd = 5


# Enum of possible unary operations.
class UnOp(Enum):
    UAdd = 1
    USub = 2
    sin = 3
    cos = 4
    tg = 5
    ctg = 6
    exp = 7
    abs = 8
    arcsin = 9
    arccos = 10
    arctg = 11
    sinh = 12
    cosh = 13
    tgh = 14
    ctgh = 15
    arcctg = 16
    ln = 17
    log2 = 18
    log10 = 19
    arcsinh = 20
    arccosh = 21
    arctgh = 22


class VectorUnOp(Enum):
    USub = 1


# Correspondence between custom and ast available binary operations
ast_bin_ops = {
    'Add': BinOp.Add,
    'Sub': BinOp.Sub,
    'Mult': BinOp.Mult,
    'Div': BinOp.Div,
    'Pow': BinOp.Pow,
    'log': BinOp.log,
    'mod': BinOp.mod,
    'vAdd': VectorBinOp.Add,
    'vSub': VectorBinOp.Sub,
    'vMulNum': VectorBinOp.MultByNum,
    'vDP': VectorBinOp.DotProd,
    'vGet': VectorBinOp.GetInd
}

functions = {
    UnOp.sin,
    UnOp.cos,
    UnOp.tg,
    UnOp.ctg,
    UnOp.exp,
    UnOp.abs,
    UnOp.sinh,
    UnOp.cosh,
    UnOp.tgh,
    UnOp.ctgh,
    UnOp.arcsin,
    UnOp.arccos,
    UnOp.arctg,
    UnOp.arcctg,
    UnOp.ln,
    UnOp.log2,
    UnOp.log10,
    BinOp.log,
    BinOp.mod,
    UnOp.arcsinh,
    UnOp.arccosh,
    UnOp.arctgh,
    VectorBinOp.Add,
    VectorBinOp.Sub,
    VectorBinOp.MultByNum,
    VectorBinOp.DotProd,
    VectorUnOp.USub
}



# Correspondence between custom and ast available unary operations
ast_un_ops = {
    'UAdd': UnOp.UAdd,
    'USub': UnOp.USub,
    'sin': UnOp.sin,
    'cos': UnOp.cos,
    'tg': UnOp.tg,
    'ctg': UnOp.ctg,
    'exp': UnOp.exp,
    'abs': UnOp.abs,
    'sinh': UnOp.sinh,
    'cosh': UnOp.cosh,
    'tgh': UnOp.tgh,
    'ctgh': UnOp.ctgh,
    'arcsin': UnOp.arcsin,
    'arccos': UnOp.arccos,
    'arctg': UnOp.arctg,
    'arcctg': UnOp.arcctg,
    'ln': UnOp.ln,
    'log2': UnOp.log2,
    'log10': UnOp.log10,
    'arcsinh': UnOp.arcsinh,
    'arccosh': UnOp.arccosh,
    'arctgh': UnOp.arctgh,
    'vUSub': VectorUnOp.USub,
}


oper_symbols = {
    BinOp.Add: '+',
    BinOp.Sub: '-',
    BinOp.Mult: '*',
    BinOp.Div: '/',
    BinOp.Pow: '**',
    UnOp.UAdd: '+',
    UnOp.USub: '-',
    UnOp.sin: 'sin',
    UnOp.cos: 'cos',
    UnOp.tg: 'tg',
    UnOp.ctg: 'ctg',
    UnOp.exp: 'exp',
    UnOp.abs: 'abs',
    UnOp.sinh: 'sinh',
    UnOp.cosh: 'cosh',
    UnOp.tgh: 'tgh',
    UnOp.ctgh: 'ctgh',
    UnOp.arcsin: 'arcsin',
    UnOp.arccos: 'arccos',
    UnOp.arctg: 'arctg',
    UnOp.arcctg: 'arcctg',
    UnOp.ln: 'ln',
    UnOp.log2: 'log2',
    UnOp.log10: 'log10',
    BinOp.log: 'log',
    BinOp.mod: 'mod',
    UnOp.arcsinh: 'arcsinh',
    UnOp.arccosh: 'arccosh',
    UnOp.arctgh: 'arctgh',
    VectorBinOp.Add: 'vAdd',
    VectorBinOp.Sub: 'vSub',
    VectorBinOp.MultByNum: 'vMulNum',
    VectorBinOp.DotProd: 'vDP',
    VectorUnOp.USub: 'vUSub'
}


def raise_(ex):
    raise ex

oper_actions = {
    BinOp.Add: lambda x, y: x + y,
    BinOp.Sub: lambda x, y: x - y,
    BinOp.Mult: lambda x, y: x * y,
    BinOp.Div: lambda x, y: x / y if y != 0 else raise_(ZeroDivisionError),
    BinOp.Pow: lambda x, y: x ** y,
    UnOp.sin: lambda x: math.sin(x),
    UnOp.cos: lambda x: math.cos(x),
    UnOp.abs: lambda x: math.fabs(x),
    UnOp.tg: lambda x: math.tan(x),
    UnOp.ctg: lambda x: 1.0/math.tan(x),
    UnOp.exp: lambda x: math.exp(x),
    UnOp.sinh: lambda x: math.sinh(x),
    UnOp.cosh: lambda x: math.cosh(x),
    UnOp.tgh: lambda x: math.tanh(x),
    UnOp.ctgh: lambda x: 1.0/math.tanh(x),
    UnOp.arcsin: lambda x: math.asin(x),
    UnOp.arccos: lambda x: math.acos(x),
    UnOp.arctg: lambda x: math.atan(x),
    UnOp.arcctg: lambda x: math.pi/2-math.atan(x),
    UnOp.arcsinh: lambda x: math.asinh(x),
    UnOp.arccosh: lambda x: math.acosh(x),
    UnOp.arctgh: lambda x: math.atanh(x),
    UnOp.log2: lambda x: math.log2(x),
    UnOp.log10: lambda x: math.log10(x),
    UnOp.ln: lambda x: math.log(x),
    BinOp.log: lambda x, y: math.log(x, y),
    BinOp.mod: lambda x, y: x % y
}


def VectorBinOp_Add(operand1, operand2):
    from sqc.src.Node import Node, NodeValue
    if operand1.shape != operand2.shape:
        raise ValueError('Vector addition shape error')
    new_elem_vector = []
    for i in range(len(operand1.vector_elements)):
        node1 = operand1.vector_elements[i]
        node2 = operand2.vector_elements[i]

        if node1.value.operand.shape != node2.value.operand.shape:
            raise ValueError('Vector addition shape error')

        if node1.value.operand.is_vector():
            new_elem_vector.append(VectorBinOp_Add(node1.value.operand, node2.value.operand))
        elif node1.value.operand.is_numeric_value():
            new_elem_vector.append(
                Node(
                    NodeValue(
                        operand=Operand(
                            numeric_value=node1.value.operand.numeric_value + node2.value.operand.numeric_value
                        )
                    )
                )
            )
    return Node(NodeValue(operand=Operand(vector_elements=new_elem_vector)))
oper_actions[VectorBinOp.Add] = VectorBinOp_Add


def VectorBinOp_Sub(operand1, operand2):
    from sqc.src.Node import Node, NodeValue
    if operand1.shape != operand2.shape:
        raise ValueError('Vector subtraction shape error')
    new_elem_vector = []
    for i in range(len(operand1.vector_elements)):
        node1 = operand1.vector_elements[i]
        node2 = operand2.vector_elements[i]

        if node1.value.operand.shape != node2.value.operand.shape:
            raise ValueError('Vector subtraction shape error')

        if node1.value.operand.is_vector():
            new_elem_vector.append(VectorBinOp_Sub(node1.value.operand, node2.value.operand))
        elif node1.value.operand.is_numeric_value():
            new_elem_vector.append(
                Node(
                    NodeValue(
                        operand=Operand(
                            numeric_value=node1.value.operand.numeric_value - node2.value.operand.numeric_value
                        )
                    )
                )
            )
    return Node(NodeValue(operand=Operand(vector_elements=new_elem_vector)))
oper_actions[VectorBinOp.Sub] = VectorBinOp_Sub


def VectorBinOp_MultByNum(operand1, operand2):
    from sqc.src.Node import Node, NodeValue
    new_elem_vector = []
    for i in range(len(operand2.vector_elements)):
        node = operand2.vector_elements[i]

        if node.value.operand.is_vector():
            new_elem_vector.append(VectorBinOp_MultByNum(operand1, node.value.operand))
        elif node.value.operand.is_numeric_value():
            new_elem_vector.append(
                Node(
                    NodeValue(
                        operand=Operand(
                            numeric_value=operand1.numeric_value * node.value.operand.numeric_value
                        )
                    )
                )
            )
    return Node(NodeValue(operand=Operand(vector_elements=new_elem_vector)))
oper_actions[VectorBinOp.MultByNum] = VectorBinOp_MultByNum


def VectorBinOp_DotProd(operand1, operand2):
    from sqc.src.Node import Node, NodeValue
    print(operand1.shape)
    if operand1.shape != operand2.shape or operand1.shape[0] != 1:
        raise ValueError('Vector dot product shape error')
    val = 0
    for i in range(len(operand1.vector_elements)):
        node1 = operand1.vector_elements[i]
        node2 = operand2.vector_elements[i]

        if node1.value.operand.shape != node2.value.operand.shape:
            raise ValueError('Vector dot product shape error')

        if node1.value.operand.is_numeric_value():
            val += node1.value.operand.numeric_value * node2.value.operand.numeric_value
    return Node(NodeValue(operand=Operand(numeric_value=val)))
oper_actions[VectorBinOp.DotProd] = VectorBinOp_DotProd


def VectorBinOp_GetInd(operand1, operand2):
    if not (0 <= operand2.numeric_value - 1 < len(operand1.vector_elements)):
        raise ValueError('Wrong index parameter')
    return copy.copy(operand1.vector_elements[operand2.numeric_value - 1])
oper_actions[VectorBinOp.GetInd] = VectorBinOp_GetInd


def VectorUnOp_USub(operand):
    return VectorBinOp_MultByNum(Operand(numeric_value=-1), operand)
oper_actions[VectorUnOp.USub] = VectorUnOp_USub



import argparse
import os

from sqc.src.Tree import Tree
from sqc.test_package import test

from sqc.src.interface import Interface


def main():
    parser = argparse.ArgumentParser(description='Symbolic quantum expression shortener')
    parser.add_argument('-t', action='store_true', help='do unit test')
    parser.add_argument('--exp', help='Input expression without spaces')
    parser.add_argument('--file', help='Input path to the file. Expression in file must be written without spaces')
    parser.add_argument('--advanced', help='Advanced mode. Only if you know what is it.')
    # parser.add_argument('--rules', help='Edit rules mode.')
    args = parser.parse_args()

    if args.t:
        test.run_tests()

    if args.exp is not None:
        is_advanced = False
        run(args.exp, is_advanced)
    elif args.advanced is not None:
        is_advanced = True
        run(args.advanced, is_advanced)
    elif args.file is not None:
        is_advanced = False
        expression = Interface()
        expression.read_from_file(args.file, is_advanced)
        if expression is not None:
            run(expression, is_advanced)


if __name__ == '__main__':
    main()


def run(st, is_advanced):
    log = open('log.txt', 'w')
    tree = Tree(expression=st)

    # log.write('Tree')
    # log.write(tree)
    tree.tree_calculate()
    print('Partially calculated expression:', tree.to_expression())
    # print('Tree walk and applying rules:')
    log.write('Partially calculated expression: ' + tree.to_expression())
    expression = Interface(tree=tree)
    node = tree.root
    step = 0
    while node is not None:
        step += 1
        print()
        print('STEP #' + str(step))
        log.write('\n STEP #'+str(step) + ' ')

        tree.tree_calculate()

        print('Tree:')
        print(tree)

        if not is_advanced:
            print('Expression before applying rule on this step:', tree.to_expression())
        else:
            print('>', tree.to_expression())

        log.write(tree.to_expression())

        exists_rule = expression.write_possible_rules(tree, is_advanced)
        if exists_rule:
            tree = expression.applying_rule(log, is_advanced)
            node = tree.root
        else:
            node = tree.next_node()



import os
import sys

from sqc.src.Tree import Tree, Rules


class Interface:
    def __init__(self, tree=None):
        self.trees_on_steps = []
        if tree is not None:
            self.trees_on_steps.append(tree)
            self.step = 0
        else:
            self.step = -1

        self.data = {}
        self.max_rules = 0

    @staticmethod
    def read_from_file(file_path, is_advanced):
        if not is_advanced:
            print("Enter 'exit' command to exit")

        while not os.path.exists(file_path):
            if file_path == "exit":
                return None

            print("can not open file. Try again")
            file_path = input()

        file = open(file_path, 'r')
        expression = file.read()
        file.close()

        """Alternative method

        ok = False
        while (not ok):
            try:
                file = open(file_path, 'r')
            except IOError as e:
                if file_path == "exit":
                    return None

                print("can not open file. Try again")
                file_path = input()
            else:
                expression = file.read()
                file.close()
                ok = True
        """
        return expression

    '''
    def write_expression(self, expression):
        if not is_advanced:
            print("Enter 'exit' command to exit")

        print("Do you want to read the expression from the file? (y/n)")

        # if we will need more variants i'll chose this interface
        #
        # print("Press 1-3")
        # print("1. read expression from file")
        # print("2. read expression from console")
        # print("3. exit")

        ok = False
        while not ok:
            answer = input().lower()
            if answer == 'y' or answer == 'yes':
                print("Enter the path to your file:")
                file_path = input()

                while not os.path.exists(file_path):
                    print("path does not exists. Try again")
                    file_path = input()

                file = open(file_path, 'w')
                file.write(expression)
                file.close()

                """Alternative method

                ok1 = False
                while (not ok1):
                    try:
                        file = open(file_path, 'r')
                    except IOError as e:
                        print("path does not exists. Try again")
                        file_path = input()
                    else:
                        file.write(expression)
                        file.close()
                        ok1 = True
                """
                ok = True
            elif answer == "n" or answer == "no":
                print(expression)
                ok = True
            elif answer == "exit":
                # ok = True
                return
            elif os.path.exists(answer):
                file = open(answer, 'w')
                file.write(expression)
                file.close()
                ok = True
            else:
                print("Error! Try again")

    def exp_to_ast(self, expression):
        return ast.parse(expression)

    def ast_to_exp(self, _ast):
        return ast.dump(_ast)

    def tree_to_exp(self, tree):
        return tree.__str__()
    '''

    def write_possible_rules(self, tree, is_advanced):
        rules = Rules()
        if not is_advanced:
            print("Choose the rule:")
        if self.step > 0:
            print('-1. Back')
        print('0. Exit')
        possible_rules = rules.get_possible_rules(tree.root, tree)
        print()
        num = 1
        for tree, rule in possible_rules:
            self.data[num] = (tree, rule)
            if not is_advanced:
                print(str(num) + '. Rule:', rule[0].to_expression(), '=', rule[1].to_expression())
                print('Expression after applying this rule:', tree.to_expression())
            else:
                print(str(num) + '. >', rule[0].to_expression(), '=', rule[1].to_expression(), '|',
                      tree.to_expression())
            num += 1

        self.max_rules = num-1
        return num > 0

    def applying_rule(self, log, is_advanced):
        ok = False
        while not ok:
            answer = input()
            try:
                answer = int(answer)
            except IOError:
                print('Error. Try again')
            else:
                if answer == 0:
                    print('Step-by-step solution saved in file log.txt')
                    log.close()
                    sys.exit()
                if answer == -1 and self.step > 0:
                    self.trees_on_steps.remove(self.trees_on_steps[self.step])
                    self.step -= 1
                    return self.trees_on_steps[self.step]
                elif 0 < answer <= self.max_rules:
                    current_tree, current_rule = self.data[answer]
                    if not is_advanced:
                        print('Leave your note for this step: ')
                    else:
                        print('note: ')

                    note = input()
                    if note is not None:
                        log.write('  # ' + note)
                    self.trees_on_steps.append(current_tree)
                    self.step += 1
                    return current_tree
                else:
                    print('There are only', self.max_rules, 'rules. Try again.')



import copy

from sqc.src.Operand import Operand

from sqc.src.operations import BinOp, UnOp, VectorUnOp, VectorBinOp
from sqc.src.operations import oper_actions


class Node:
    """
    Node class.
        value:      NodeValue object
        children:   list of outgoing nodes
    """

    def __init__(self, value):
        self.value = value
        self.children = []
        self.parent = None

    def __eq__(self, other):
        return self.value == other.value and len(self.children) == len(other.children)

    def __str__(self):
        return str(self.value)

    def __copy__(self):
        result = Node(copy.copy(self.value))
        for child in self.children:
            result.add_child(copy.copy(child))
        return result

    def add_child(self, node):
        node.parent = self
        self.children.append(node)

    def eq_subtrees(self, other):
        """
        Recursive function for checking whether subtrees are equal.

        :param other:   another node
        :type other:    Node
        :return:        True if corresponding subtrees are equal, False o/w.
        :rtype:         bool
        """

        result = self == other
        if result:
            for i in range(len(self.children)):
                child = self.children[i]
                other_child = other.children[i]

                result &= child.eq_subtrees(other_child)
                if not result:
                    break
        return result

    def calculate(self):
        """
        Tries to calculate children. (for binary trees only)
        :return:
        :rtype:
        """
        value_children = \
            [child for child in self.children if child.value.is_operand() and child.value.operand.is_numeric_value()]
        if self.value.is_operation() and isinstance(self.value.operation, BinOp) and \
           len(value_children) == 2:
            values = [child.value.operand.numeric_value for child in self.children]
            value = oper_actions[self.value.operation](values[0], values[1])

            self.children = []
            self.value = NodeValue(operand=Operand(numeric_value=value))

        elif self.value.is_operation() and isinstance(self.value.operation, UnOp) and len(value_children) == 1:
            values = [child.value.operand.numeric_value for child in self.children]
            if self.value.operation == UnOp.UAdd:
                self.children = []
                self.value = NodeValue(operand=Operand(numeric_value=values[0]))
            elif self.value.operation == UnOp.USub:
                self.children = []
                self.value = NodeValue(operand=Operand(numeric_value=-values[0]))
            else:
                value = oper_actions[self.value.operation](values[0])

                self.children = []
                self.value = NodeValue(operand=Operand(numeric_value=value))

        elif self.value.is_operation() and (isinstance(self.value.operation, VectorBinOp)):
            is_mult_by_num = False
            is_get_ind = False
            args_are_numeric_values = True
            if self.value.operation == VectorBinOp.MultByNum:
                is_mult_by_num = True
                if not self.children[0].value.operand.is_numeric_value():
                    self.children[0], self.children[1] = self.children[1], self.children[0]
                if not self.children[0].value.operand.is_numeric_value():
                    raise ValueError('Wrong multiplication parameter')
            if self.value.operation == VectorBinOp.GetInd:
                is_get_ind = True
                if not self.children[1].value.operand.is_numeric_value():
                    raise ValueError('Wrong get index parameter')

            for i in range(int(is_mult_by_num), 2 - int(is_get_ind)):
                if self.children[i].value.is_operand():
                    if self.children[i].value.operand.is_vector():
                        for node in self.children[i].value.operand.vector_elements:
                            if node.value.operand.is_vector():
                                for node2 in node.value.operand.vector_elements:
                                    if not node2.value.operand.is_numeric_value():
                                        args_are_numeric_values = False
                                        break
                                if not args_are_numeric_values:
                                    break
                            else:
                                if not node.value.operand.is_numeric_value():
                                    args_are_numeric_values = False
                                    break
                    elif not self.children[i].value.operand.is_variable():
                        raise ValueError('Binary vector operation arguments error')
                    else:
                        args_are_numeric_values = False
                        break
                else:
                    args_are_numeric_values = False
                    break

            if args_are_numeric_values:
                self.value = oper_actions[self.value.operation](
                    self.children[0].value.operand,
                    self.children[1].value.operand
                ).value
                self.children = []

        elif self.value.is_operation() and (isinstance(self.value.operation, VectorUnOp)):
            args_are_numeric_values = True
            if self.children[0].value.is_operand():
                if self.children[0].value.operand.is_vector():
                    for node in self.children[0].value.operand.vector_elements:
                        if node.value.operand.is_vector():
                            for node2 in node.value.operand.vector_elements:
                                if not node2.value.operand.is_numeric_value():
                                    args_are_numeric_values = False
                                    break
                            if not args_are_numeric_values:
                                break
                        else:
                            if not node.value.operand.is_numeric_value():
                                args_are_numeric_values = False
                                break
                elif not self.children[0].value.operand.is_variable():
                    raise ValueError('Binary vector operation arguments error')
                else:
                    args_are_numeric_values = False
            else:
                args_are_numeric_values = False

            if args_are_numeric_values:
                self.value = oper_actions[self.value.operation](self.children[0].value.operand).value
                self.children = []

    def unify(self, rule_node):
        if self.value is None:
            return None

        if self.value.is_operand() and not rule_node.value.is_operand():
            return None

        result_data = {}
        if self.value.is_operation():
            if rule_node.value.is_operation():
                if len(self.children) != len(rule_node.children) or\
                   self.value.operation != rule_node.value.operation:
                    return None
                for i in range(len(self.children)):
                    inter_result = self.children[i].unify(rule_node.children[i])

                    if inter_result is None:
                        return None

                    for key, value in inter_result.items():
                        if key in result_data:
                            if not value.eq_subtrees(result_data[key]):
                                return None
                        else:
                            result_data[key] = value
            else:
                if not rule_node.value.operand.is_variable() or rule_node.value.operand.is_constant():
                    return None
                else:
                    result_data[rule_node.value.operand.variable] = copy.copy(self)
        else:
            if rule_node.value.operand.is_numeric_value() or rule_node.value.operand.is_constant():
                if self.value != rule_node.value:
                    return None
            elif not self.value.operand.is_vector():
                result_data[rule_node.value.operand.variable] = copy.copy(self)
            else:
                return None

        return result_data

    def replace(self, rule):
        substitutions = self.unify(rule[0].root)

        def dfs(rule_node, parent, substitutions):
            if rule_node.value.is_operand():
                if rule_node.value.operand.is_variable() and not rule_node.value.operand.is_constant():
                    new_node = copy.copy(substitutions[rule_node.value.operand.variable])
                else:
                    new_node = copy.copy(rule_node)
                parent.add_child(new_node)
            else:
                new_node = Node(copy.copy(rule_node.value))
                parent.add_child(new_node)

                for i in range(len(rule_node.children)):
                    dfs(rule_node.children[i], new_node, substitutions)

        tmp_root = Node(value=None)
        dfs(rule[1].root, tmp_root, substitutions)

        new_root = tmp_root.children[0]
        self.value = new_root.value
        self.children = new_root.children


class NodeValue:
    """
    NodeValue can be whether an operation or operand.
    """

    def __init__(self, operation=None, operand=None):
        """

        :param operation:
        :type operation:
        :param operand:
        :type operand:
        """

        if operation is not None:
            self.type = 0
            self.operation = operation
        else:
            self.type = 1
            self.operand = operand

    def __str__(self):
        if self.is_operand():
            return str(self.operand)
        else:
            return str(self.operation.name)

    def __eq__(self, other):
        return self.is_operation() and other.is_operation() and self.operation == other.operation or\
                self.is_operand() and other.is_operand() and self.operand == other.operand

    def __copy__(self):
        if self.is_operation():
            return NodeValue(operation=self.operation)
        else:
            return NodeValue(operand=copy.copy(self.operand))

    def is_operation(self):
        return self.type == 0

    def is_operand(self):
        return self.type == 1



import copy
from sqc.src.Constant import constants


class Operand:
    """
    Operand can be whether a numeric value or variable of given dimension.
    """
    def __init__(self, numeric_value=None, variable=None, vector_elements=None):
        """
        :param numeric_value:   Number/vector/matrix
        :type numeric_value:    int, float
        :param variable:        Name of the variable or list/matrix of variable.
        :type variable:         string
        :param vector_elements: List of nodes
        :type vector_elements:  list
        """

        self._is_constant = False
        self.shape = None
        if numeric_value is not None:
            self.shape = ()
            self.type = 0
            self.numeric_value = numeric_value
        elif variable is not None:
            self.type = 1
            self.variable = variable
            if self.variable in constants.keys():
                self._is_constant = True
        else:
            self.type = 2
            self.vector_elements = vector_elements

            elem_shape = None
            for node in self.vector_elements:
                if node.value.is_operand():
                    if node.value.operand.is_variable() and not node.value.operand.is_constant():
                        elem_shape = None
                        break
                    elif node.value.operand.is_numeric_value() or node.value.operand.is_constant():
                        if elem_shape is None:
                            elem_shape = ()
                        elif elem_shape != ():
                            raise ValueError('Vector elements shape error')
                    elif node.value.operand.is_vector():
                        if elem_shape is None:
                            elem_shape = (len(node.value.operand.vector_elements), )
                        elif elem_shape != (len(node.value.operand.vector_elements), ):
                            raise ValueError('Vector elements shape error')
                else:
                    raise ValueError('Operation in vector element is not allowed yet')

            if elem_shape is not None:
                if elem_shape == ():
                    self.shape = (1, len(self.vector_elements))
                else:
                    self.shape = (len(self.vector_elements), elem_shape[0])

    def __str__(self):
        if self.is_numeric_value():
            return str(self.numeric_value)
        elif self.is_variable():
            return str(self.variable)
        else:
            result = '['
            for it in self.vector_elements:
                result += str(it) + ', '
            if len(result) > 1:
                return result[0:-2] + ']'
            else:
                return '[]'

    def __eq__(self, other):
        if self.is_variable():
            return other.is_variable() and self.variable == other.variable
        elif self.is_numeric_value():
            return other.is_numeric_value() and self.numeric_value == other.numeric_value
        else:
            if other.is_vector() and len(other.vector_elements) == len(self.vector_elements):
                eq = True
                for i in range(len(self.vector_elements)):
                    eq = self.vector_elements[i].eq_subtrees(other.vector_elements[i])
                    if not eq:
                        break
                return eq
            else:
                return False

    def __copy__(self):
        if self.is_numeric_value():
            return Operand(numeric_value=self.numeric_value)
        elif self.is_variable():
            return Operand(variable=self.variable)
        else:
            vector_elements = []
            for it in self.vector_elements:
                vector_elements.append(copy.copy(it))
            return Operand(vector_elements=vector_elements)

    def is_numeric_value(self):
        return self.type == 0

    def is_variable(self):
        return self.type == 1

    def is_vector(self):
        return self.type == 2

    def is_constant(self):
        return self._is_constant



import unittest

from sqc.src.Tree import Tree


class Test(unittest.TestCase):
    def test_empty(self):
        with self.assertRaises(Exception):
            Tree(expression='[]')

    def test_uni_vector(self):
        input_string = '[1]'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[1]')

    def test_empty_double_vector(self):
        input_string = '[[2], [1]]'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[2], [1]]')

    def test_vector1(self):
        input_string = '[x]'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[x]')

    def test_vector2(self):
        input_string = '[x, y]'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[x, y]')

    def test_vector3(self):
        input_string = '[[1, x], [2, y]]'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[1, x], [2, y]]')

    def test_vector_ops_Add_1(self):
        input_string = 'vAdd([3], [1])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[4]')

    def test_vector_ops_Add_2(self):
        input_string = 'vAdd([3, 4], [1, 6])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[4, 10]')

    def test_vector_ops_Add_3(self):
        input_string = 'vAdd([[3]], [[1]])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[4]]')

    def test_vector_ops_Add_4(self):
        input_string = 'vAdd([[3, 2]], [[1, 5]])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[4, 7]]')

    def test_vector_ops_Add_5(self):
        input_string = 'vAdd([[3, 5], [1, 6], [3, 7]], [[3, 5], [1, 6], [3, 7]])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[6, 10], [2, 12], [6, 14]]')

    def test_vector_ops_Sub_1(self):
        input_string = 'vSub([3], [1])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[2]')

    def test_vector_ops_Sub_2(self):
        input_string = 'vSub([3, 4], [1, 6])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[2, -2]')

    def test_vector_ops_Sub_3(self):
        input_string = 'vSub([[3]], [[1]])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[2]]')

    def test_vector_ops_Sub_4(self):
        input_string = 'vSub([[3, 2]], [[1, 5]])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[2, -3]]')

    def test_vector_ops_Sub_5(self):
        input_string = 'vSub([[3, 5], [1, 6], [3, 7]], [[3, 5], [1, 6], [3, 7]])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[0, 0], [0, 0], [0, 0]]')

    def test_vector_ops_MulNum_1(self):
        input_string = 'vMulNum(3, [1])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[3]')

    def test_vector_ops_MulNum_2(self):
        input_string = 'vMulNum([3, 4], -2)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[-6, -8]')

    def test_vector_ops_MulNum_3(self):
        input_string = 'vMulNum(1, [[3], [1]])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[3], [1]]')

    def test_vector_ops_MulNum_4(self):
        input_string = 'vMulNum(-2, [[3, 2], [1, 5]])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[-6, -4], [-2, -10]]')

    def test_vector_ops_MulNum_5(self):
        input_string = 'vMulNum([1], [[3, 2], [1, 5]])'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            tree.tree_calculate()

    def test_vector_ops_DP_1(self):
        input_string = 'vDP([3], [1])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '3')

    def test_vector_ops_DP_2(self):
        input_string = 'vDP([3, 4], [1, 6])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '27')

    def test_vector_ops_DP_3(self):
        input_string = 'vDP([3, 4], [1, 6])-27'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '0')

    def test_vector_ops_DP_4(self):
        input_string = 'vDP([3, 4], [1, 6])/10'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '2.7')

    def test_vector_ops_DP_5(self):
        input_string = 'vDP([1], [[3, 2], [1, 5]])'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            tree.tree_calculate()

    def test_vector_ops_DP_6(self):
        input_string = 'vDP([1], [3, 2])'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            tree.tree_calculate()

    def test_vector_ops_Get_1(self):
        input_string = 'vGet([3], 1)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '3')

    def test_vector_ops_Get_2(self):
        input_string = 'vGet([3, 4], 2)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '4')

    def test_vector_ops_Get_3(self):
        input_string = 'vGet([[3], [4]], 1)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[3]')

    def test_vector_ops_Get_4(self):
        input_string = 'vGet([[3, 4], [1, 6]], 2)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[1, 6]')

    def test_vector_ops_Get_5(self):
        input_string = 'vGet(1, [[3, 2], [1, 5]])'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            tree.tree_calculate()

    def test_vector_ops_Get_6(self):
        input_string = 'vGet([3, 2], 1.0)'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            tree.tree_calculate()

    def test_vector_ops_Get_7(self):
        input_string = 'vGet([3, 2], 3)'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            tree.tree_calculate()

    def test_vector_ops_USub_1(self):
        input_string = 'vUSub([[3], [4]])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[-3], [-4]]')

    def test_vector_ops_USub_2(self):
        input_string = 'vUSub([1])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[-1]')

    def test_vector_ops_USub_3(self):
        input_string = 'vUSub(1)'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            tree.tree_calculate()

if __name__ == '__main__':
    unittest.main()
